module Transpiler
using CUDA
using ..ExpressionProcessing

# Number of threads per block/SM + max number of registers
# https://docs.nvidia.com/cuda/cuda-c-programming-guide/#features-and-technical-specifications
# Need to assume a max of 2048 threads per Streaming Multiprocessor (SM)
# One SM can have 64*1024 32-bit registers at max
# One thread can at max use 255 registers
# Meaning one has access to at most 32 registers in the worst case. Using 64 bit values this number gets halfed (see: https://docs.nvidia.com/cuda/cuda-c-programming-guide/#multiprocessor-level (almost at the end of the linked section))

# Maybe helpful for future performance tuning: https://docs.nvidia.com/cuda/cuda-c-programming-guide/#maximum-number-of-registers-per-thread

# https://docs.nvidia.com/cuda/cuda-c-programming-guide/#multiprocessor-level
# This states, that using fewer registers allows more threads to reside on a single SM which improves performance. 
# So I could use more registers at the expense for performance. Depending on how this would simplify my algorithm, I might do this and leave more optimisation to future work

# Since the generated expressions should have between 10 and 50 symbols, I think allowing a max. of 128 32-bit registers should make for an easy algorithm. If during testing the result is slow, maybe try reducing the number of registers and perform more intelligent allocation/assignment
# With 128 Registers, one could have 32 Warps on one SM ((128 * 16 = 2048) * 32 == 64*1024 == max number of registers per SM) This means 512 Threads per SM in the worst case

#
# Make a "function execute(...)" that takes the data and the transpiled code. Pass the data to the kernel and start executing
# Note: Maybe make an additional function that transpiles and executed the code. This would then be the function the user calls
#

# To increase performance, it would probably be best for all helper functions to return their IO Buffer and not a string
function transpile(expression::ExpressionProcessing.PostfixType)::String
	exitJumpLocationMarker = "\$L__BB0_2"
	ptxBuffer = IOBuffer()

	println(ptxBuffer, get_cuda_header())
	println(ptxBuffer, get_kernel_signature("ExpressionProcessing", [Int32, Float32]))
	println(ptxBuffer, "{")

	# TODO: Parameter loading

	# TODO: once parameters are loaded, the second parameter for the guard clause can be set
	temp = get_next_free_register("r")
	guardClause = get_guard_clause(exitJumpLocationMarker, temp) # since we need to know how many registers we used, we cannot yet write the guard clause to the ptxBuffer

	calc_code = generate_calculation_code(expression)
	println(ptxBuffer, get_register_definitions())
	println(ptxBuffer, guardClause)
	println(ptxBuffer, calc_code)

	# exit jump location
	print(ptxBuffer, exitJumpLocationMarker); println(ptxBuffer, ": ret;")
	println(ptxBuffer, "}")

	generatedCode = String(take!(ptxBuffer))
	println(generatedCode)
	return generatedCode
end

# TODO: Make version, target and address_size configurable; also see what address_size means exactly
function get_cuda_header()::String
	return "
.version 7.1
.target sm_52
.address_size 64
"
end

function get_kernel_signature(kernelName::String, parameters::Vector{DataType})::String
	signatureBuffer = IOBuffer()
	print(signatureBuffer, ".visible .entry ")
	print(signatureBuffer, kernelName)
	println(signatureBuffer, "(")

	
	for i in eachindex(parameters)
		type = type_to_ptx_type(parameters[i])
		print(signatureBuffer, 
			  "  .param ", type, " ", kernelName, "_param_", i)
		if i != lastindex(parameters)
			println(signatureBuffer, ",")
		end
	end

	print(signatureBuffer, ")")
	return String(take!(signatureBuffer))
end

"
Constructs the PTX code used for handling the case where too many threads are started.

- param ```nrOfVarSetsRegister```: The register which holds the total amount of variable sets for the kernel
"
function get_guard_clause(exitJumpLocation::String, nrOfVarSetsRegister::String)::String
	guardBuffer = IOBuffer()

	threadIds = get_next_free_register("r")
	threadsPerCTA = get_next_free_register("r")
	currentThreadId = get_next_free_register("r")

	# load data into above defined registers
	println(guardBuffer, "mov.u32    $threadIds, %ntid.x;")
	println(guardBuffer, "mov.u32    $threadsPerCTA, %ctaid.x;")
	println(guardBuffer, "mov.u32    $currentThreadId, %tid.x;")

	globalThreadId = get_next_free_register("r") # basically the index of the thread in the variable set
	breakCondition = get_next_free_register("p")
	println(guardBuffer, "mad.lo.s32    $globalThreadId, $threadIds, $threadsPerCTA, $currentThreadId;")
	println(guardBuffer, "setp.ge.s32    $breakCondition, $globalThreadId, $nrOfVarSetsRegister;") # guard clause = index > nrOfVariableSets

	# branch to end if breakCondition is true
	print(guardBuffer, "@$breakCondition bra    $exitJumpLocation;")

	return String(take!(guardBuffer))
end

# Current assumption: Expression only made out of constant values
function generate_calculation_code(expression::ExpressionProcessing.PostfixType)::String
	codeBuffer = IOBuffer()
	operands = Vector{Union{Float32, String}}()

	println(expression)
	for i in eachindex(expression)
		token = expression[i]

		if token.Type == FLOAT32
			push!(operands, reinterpret(Float32, token.Value))
		elseif token.Type == OPERATOR
			# get_ptx_operator will be reworked completly to return the code for the operation
			operator = get_ptx_operator(reinterpret(Operator, token.Value))
			register = get_next_free_register("f")
			print(codeBuffer, "    $operator $register, ")

			ops = last(operands, 2)
			pop!(operands);pop!(operands)
			print(codeBuffer, join(ops, ", "))
			println(codeBuffer, ";")

			push!(operands, register)
		elseif token.Type == INDEX
			# TODO
		end
	end

	return String(take!(codeBuffer))
end

function type_to_ptx_type(type::DataType)::String
	if type == Int64
		return ".s64"
	elseif type == Float32
		return ".f32"
	else
		return ".b64"
	end
end

# TODO: Change this, to return the entire calculation not just the operator. Because for POWER and EXP we need multiple instructions to calculate them (seperation of concerns).
function get_ptx_operator(operator::Operator)::String
	if operator == ADD
		return "add.f32"
	elseif operator == SUBTRACT
		return "sub.f32"
	elseif operator == MULTIPLY
		return "mul.f32"
	elseif operator == DIVIDE
		return "div.approx.f32"
	elseif operator == POWER
		return ""
	elseif operator == ABS
		return "abs.f32"
	elseif operator == LOG
		return "lg2.approx.f32"
	elseif operator == EXP
		return ""
	elseif operator == SQRT
		return "sqrt.approx.f32"
	else
		throw(ArgumentError("Operator conversion to ptx not implemented for $operator"))
	end
end

let registers = Dict() # stores the count of the register already used.
	global get_next_free_register
	global get_register_definitions

	# By convention these names correspond to the following types:
	# - p -> pred
	# - f -> float32
	# - r -> 32 bit
	# - var -> float32
	# - param -> float32 !! although, they might get inserted as fixed number and not be sent to gpu?
	function get_next_free_register(name::String)::String
		if haskey(registers, name)
			registers[name] += 1
		else
			registers[name] = 1
		end

		return string("%", name, registers[name] - 1)
	end

	function get_register_definitions()::String
		registersBuffer = IOBuffer()
	
		for definition in registers
			regType = ""
			if definition.first == "p"
				regType = ".pred"
			elseif definition.first == "f"
				regType = ".f32"
			elseif definition.first == "var"
				regType = ".f32"
			elseif definition.first == "param"
				regType = ".f32"
			elseif definition.first == "r"
				regType = ".b32"
			else
				throw(ArgumentError("Unknown register name used. Name '$(definition.first)' cannot be mapped to a PTX type."))
			end
			println(registersBuffer, ".reg $regType   %$(definition.first)<$(definition.second)>;")
		end
	
		return String(take!(registersBuffer))
	end
end

end

